# -----------------------------------------------------------------------------
# Commons - Utility classes for iTaxoTools modules
# Copyright (C) 2021  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""Multithreading classes for PySide6"""

from PySide6 import QtCore

import multiprocessing
import sys
import traceback

from . import io


class FailedError(Exception):
    def __init__(self, code=1):
        super().__init__(f"Thread failed with code: {code}")
        self.code = code


class CancelledError(Exception):
    def __init__(self, code=-1):
        super().__init__(f"Thread cancelled with code: {code}")
        self.code = code


class TerminatedError(Exception):
    def __init__(self):
        super().__init__("Thread was forcibly terminated")


class WorkerThread(QtCore.QThread):
    """
    Multithreaded function execution using PySide6.
    Initialise with the function and parameters for execution.
    Signals are used to communicate with parent thread.
    Use self.start() to fork/spawn.

    Signals:
    ----------
    started():
        Inherited. Emitted on start.
    finished():
        Inherited. Emitted on finish, regardless of success.
    done(result):
        Emitted when the called function returned successfully.
        Passes the result.
    fail(exception, traceback):
        Emitted if an exception occured. Passes the exception and traceback.
    cancel(exception):
        Emitted if the thread was cancelled.
    """

    done = QtCore.Signal(object)
    fail = QtCore.Signal(Exception, str)
    cancel = QtCore.Signal(CancelledError)

    def __init__(self, function, *args, **kwargs):
        """
        Parameters
        ----------
        function : function
            The function to run on the new thread.
        *args : positional arguments, optional
            If given, they are passed on to the function.
        **kwargs : keyword arguments, optional
            If given, they are passed on to the function.
        """
        super().__init__()
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.timeout = 1000

    def run(self):
        """
        Overloads QThread. This is run on a new thread.
        Do not call this directly, call self.start() instead.
        """
        try:
            result = self.function(*self.args, **self.kwargs)
        except CancelledError as exception:
            self.cancel.emit(exception)
        except Exception as exception:
            trace = traceback.format_exc()
            self.fail.emit(exception, trace)
        else:
            self.done.emit(result)

    def check(self):
        """
        Call this from within a worker thread to check if the thread
        should exit by user request. If so, CancelledError is raised,
        which should then be handled by WorkerThread.run().
        """
        if self.isInterruptionRequested():
            raise CancelledError(-1)

    def exit(self, code: int = 0):
        """
        Call this from within a worker thread to indicate the work is done.
        Argument `code` determines transition behaviour: zero for success,
        a positive value for failure, a negative value on interruption.
        """
        super().exit(code)

    def terminate(self):
        """
        Attempt to cleanly exit, then forcibly terminate
        after a short timeout (dangerous!)
        """
        self.requestInterruption()
        self.exit(-1)
        if not self.wait(self.timeout):
            super().terminate()
            self.wait()
            self.cancel.emit(TerminatedError())


class Worker:
    """
    Used by Process to spawn a new process.
    Holds pipe information for stdio redirection.
    """

    def __getstate__(self):
        """Required for process spawning."""
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        """Required for process spawning."""
        self.__dict__ = state

    def __init__(self, uproc):
        """Configure self from given Process"""
        uproc.pipeIn[1].close()
        self.pipeControl = uproc.pipeControl[1]
        self.pipeData = uproc.pipeData[1]
        self.pipeOut = uproc.pipeOut[1]
        self.pipeErr = uproc.pipeErr[1]
        self.pipeIn = uproc.pipeIn[0]

    def target(self, function, *args, **kwargs):
        """
        This is executed as a new process.
        Alerts parent process via pipe.
        """
        out = io.PipeIO(self.pipeOut, "w")
        err = io.PipeIO(self.pipeErr, "w")
        inp = io.PipeIO(self.pipeIn, "r")

        # import sys
        sys.stdout = out
        sys.stderr = err
        sys.stdin = inp

        try:
            result = function(*args, **kwargs)
            self.pipeControl.send("RESULT")
            self.pipeData.send(result)
        except Exception as exception:
            trace = traceback.format_exc()
            self.pipeControl.send("EXCEPTION")
            self.pipeData.send((exception, trace))
        finally:
            self.pipeControl.close()
            self.pipeData.close()
            self.pipeOut.close()
            self.pipeErr.close()
            self.pipeIn.close()


class Process(QtCore.QThread):
    """
    Multiprocess function execution using PySide6.
    Unlike QtCore.QProcess, this launches python functions, not programs.
    Use self.start() to fork/spawn.

    Signals
    -------
    done(result) : On success
    fail(exception, traceback) : On exception raised
    error(exitcode) : On process exit

    Example
    -------

    def work(number):
        print('This runs on the child process.')
        print('Trying to use PyQt in here won\'t work.')
        return number * 2

    def success(result):
        print('This runs on the parent process.')
        print('You can spawn dialog messages here.')
        QMessageBox.information(None, 'Success',
            'Result = '+str(result), QMessageBox.Ok)
        print('This prints 42: ', result)

    self.process = Process(work, 21)
    self.process.done.connect(success)
    self.process.start()
    return

    """

    done = QtCore.Signal(object)
    fail = QtCore.Signal(Exception, str)
    error = QtCore.Signal(int)

    def __init__(self, function, *args, **kwargs):
        """
        Parameters
        ----------
        function : function
            The function to run on the new process.
        *args : positional argument, optional
            If given, it is passed on to the function.
        **kwargs : keyword arguments, optional
            If given, it is passed on to the function.
        """
        super().__init__()
        self._quit = None
        self._data = None
        self._data_obj = None
        self.stream = None
        self.pipeControl = multiprocessing.Pipe(duplex=True)
        self.pipeData = multiprocessing.Pipe(duplex=True)
        self.pipeOut = multiprocessing.Pipe(duplex=False)
        self.pipeErr = multiprocessing.Pipe(duplex=False)
        self.pipeIn = multiprocessing.Pipe(duplex=False)
        self.worker = Worker(self)
        self.process = multiprocessing.Process(
            target=self.worker.target,
            daemon=True,
            args=(function,) + args,
            kwargs=kwargs,
        )

    def setStream(self, stream):
        """Send process output to given file-like stream"""
        self.stream = stream

        if stream is not None:
            self.handleOut = self._streamOut
            self.handleErr = self._streamOut
        else:
            self.handleOut = self._streamNone
            self.handleErr = self._streamNone

    def _streamNone(self, data):
        pass

    def _streamOut(self, data):
        self.stream.write(data)

    def run(self):
        """
        Overloads QThread. This is run on a new thread.
        Do not call this directly, call self.start() instead.
        Starts and watches the process.
        """
        self._quit = False
        self.process.start()
        self.pipeData[1].close()
        self.pipeOut[1].close()
        self.pipeErr[1].close()
        self.pipeControl = self.pipeControl[0]
        self.pipeData = self.pipeData[0]
        self.pipeOut = self.pipeOut[0]
        self.pipeErr = self.pipeErr[0]
        self.pipeIn = self.pipeIn[1]

        sentinel = self.process.sentinel
        waitList = {
            sentinel: None,
            self.pipeControl: self.handleControl,
            self.pipeOut: self.handleOut,
            self.pipeErr: self.handleErr,
        }
        while waitList and sentinel is not None:
            readyList = multiprocessing.connection.wait(waitList.keys())
            for pipe in readyList:
                if pipe == sentinel:
                    # Process exited, but must make sure
                    # all other pipes are empty before quitting
                    if len(readyList) == 1:
                        sentinel = None
                else:
                    try:
                        data = pipe.recv()
                    except EOFError:
                        waitList.pop(pipe)
                    else:
                        waitList[pipe](data)

        # Emit the proper signal
        if self.process.exitcode != 0 and not self._quit:
            self.error.emit(self.process.exitcode)
        elif self._data == "RESULT":
            self.done.emit(self._data_obj)
        elif self._data == "EXCEPTION":
            self.fail.emit(*self._data_obj)

    def handleControl(self, data):
        """Handle control pipe signals"""
        self._data = data
        self._data_obj = self.pipeData.recv()
        self.process.join()

    def handleOut(self, data):
        """Overload this to handle process stdout"""

    def handleErr(self, data):
        """Overload this to handle process stderr"""

    def quit(self):
        """Clean exit"""
        self._quit = True
        if self.process is not None and self.process.is_alive():
            self.process.terminate()
        super().quit()



# -----------------------------------------------------------------------------
# Commons - Utility classes for iTaxoTools modules
# Copyright (C) 2021  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""Widgets for dialog windows"""


from PySide6 import QtCore, QtWidgets

import enum
import traceback

from .graphics import ScalingImage
from .layouts import VLineSeparator


class PushButton(QtWidgets.QPushButton):
    """A larger button with square borders"""

    def __init__(self, *args, **kwargs):
        onclick = None
        if "onclick" in kwargs:
            onclick = kwargs.pop("onclick")
        super().__init__(*args, **kwargs)
        if onclick:
            self.clicked.connect(onclick)
        self.setStyleSheet(
            """
            QPushButton {
                border: 1px solid Palette(Mid);
                background: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                    stop: 0 Palette(Light), stop: 1 Palette(Window));
                color: Palette(Text);
                padding: 5px 15px;
                min-width: 80px;
                outline: none;
                }
            QPushButton:hover {
                border: 1px solid qlineargradient(x1: -1, y1: 0, x2: 0, y2: 2,
                    stop: 0 Palette(Midlight), stop: 1 Palette(Mid));
                background: Palette(Light);
                }
            QPushButton:focus:hover {
                border: 1px solid Palette(Highlight);
                background: Palette(Light);
                }
            QPushButton:focus:!hover {
                border: 1px solid Palette(Highlight);
                background: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 10,
                    stop: 0 Palette(Light), stop: 1 Palette(Highlight));
                }
            QPushButton:focus:pressed {
                border: 1px solid Palette(Highlight);
                background: Palette(Midlight);
                }
            QPushButton:pressed {
                border: 1px solid Palette(Highlight);
                background: Palette(Midlight);
                }
            QPushButton:disabled {
                border: 1px solid qlineargradient(x1: 0, y1: -1, x2: 0, y2: 2,
                    stop: 0 Palette(Midlight), stop: 1 Palette(Dark));
                background: Palette(Midlight);
                color: Palette(Dark)
        }"""
        )


class _HeaderLabels(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.tool = QtWidgets.QLabel("TOOL")
        self.tool.setAlignment(QtCore.Qt.AlignBottom)
        self.tool.setStyleSheet(
            """
            color: palette(Text);
            font-size: 14px;
            letter-spacing: 1px;
            font-weight: bold;
            text-decoration: underline;
            """
        )

        self.citation = QtWidgets.QLabel("CITATION")
        self.citation.setAlignment(QtCore.Qt.AlignBottom)
        self.citation.setStyleSheet(
            """
            color: palette(Shadow);
            font-size: 12px;
            font-style: italic;
            """
        )

        self.task = QtWidgets.QLabel("TASK")
        self.task.setAlignment(QtCore.Qt.AlignBottom)
        self.task.setStyleSheet(
            """
            color: palette(Shadow);
            font-size: 14px;
            font-weight: bold;
            letter-spacing: 1px;
            """
        )

        self.description = QtWidgets.QLabel("DESCRIPTION")
        self.description.setAlignment(QtCore.Qt.AlignTop)
        self.description.setStyleSheet(
            """
            color: palette(Text);
            font-size: 12px;
            letter-spacing: 1px;
            """
        )

        layout = QtWidgets.QGridLayout()
        layout.setRowStretch(0, 2)
        layout.addWidget(self.tool, 1, 0, 1, 2)
        layout.addWidget(self.citation, 1, 2)
        layout.addWidget(self.task, 2, 0, 1, 2)
        layout.addWidget(self.description, 3, 1, 1, 3)
        layout.setRowStretch(4, 2)
        layout.setColumnStretch(3, 1)
        layout.setRowMinimumHeight(0, 6)
        layout.setRowMinimumHeight(4, 6)
        layout.setColumnMinimumWidth(0, 4)
        layout.setHorizontalSpacing(4)
        layout.setVerticalSpacing(4)
        layout.setContentsMargins(0, 0, 0, 4)
        self.setLayout(layout)


class _HeaderToolBar(QtWidgets.QToolBar):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setIconSize(QtCore.QSize(32, 32))
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum
        )
        self.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self.setStyleSheet(
            """
            QToolBar {
                spacing: 2px;
                }
            QToolButton {
                color: palette(ButtonText);
                background: transparent;
                border: 2px solid transparent;
                border-radius: 3px;
                font-size: 14px;
                min-width: 50px;
                min-height: 60px;
                padding: 6px 0px 0px 0px;
                margin: 4px 0px 4px 0px;
                }
            QToolButton:hover {
                background: palette(Window);
                border: 2px solid transparent;
                }
            QToolButton:pressed {
                background: palette(Midlight);
                border: 2px solid palette(Mid);
                border-radius: 3px;
                }
            QToolButton[popupMode="2"]:pressed {
                padding-bottom: 5px;
                border: 1px solid palette(Dark);
                margin: 5px 1px 0px 1px;
                border-bottom-right-radius: 0px;
                border-bottom-left-radius: 0px;
                }
            QToolButton::menu-indicator {
                image: none;
                width: 30px;
                border-bottom: 2px solid palette(Mid);
                subcontrol-origin: padding;
                subcontrol-position: bottom;
                }
            QToolButton::menu-indicator:disabled {
                border-bottom: 2px solid palette(Midlight);
                }
            QToolButton::menu-indicator:pressed {
                border-bottom: 0px;
                }
            """
        )


class Header(QtWidgets.QFrame):
    """
    The Taxotools toolbar, with space for a title, description,
    citations and two logos.
    """

    def __init__(self):
        super().__init__()
        self.dictTool = {"title": "", "citation": "", "description": ""}
        self.dictTask = {"title": "", "description": ""}
        self._logoTool = None
        self.draw()

    def draw(self):
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Maximum
        )
        self.setStyleSheet(
            """
            Header {
                background: palette(Light);
                border-top: 1px solid palette(Mid);
                border-bottom: 1px solid palette(Dark);
                }
            """
        )

        self.labels = _HeaderLabels()
        self.widgetLogoTool = QtWidgets.QLabel()
        self.widgetLogoTool.setAlignment(QtCore.Qt.AlignCenter)
        self.widgetLogoProject = ScalingImage()
        self.toolbar = _HeaderToolBar()
        self.widget = QtWidgets.QWidget()

        layout = QtWidgets.QHBoxLayout()
        layout.addSpacing(8)
        layout.addWidget(self.widgetLogoTool)
        layout.addSpacing(8)
        layout.addWidget(self.labels, 0)
        layout.addSpacing(8)
        layout.addWidget(self.toolbar, 0)
        layout.addWidget(self.widget, 1)
        layout.addSpacing(12)
        layout.addWidget(self.widgetLogoProject)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

    def setTool(self, title, citation, description):
        self.showTool(title=title, citation=citation, description=description)
        self.updateLabelWidth()

    def showTool(self, **kwargs):
        d = {k: v for k, v in kwargs.items() if v is not None}
        d = {k: v for k, v in d.items() if k in self.dictTool.keys()}
        self.dictTool.update(d)
        self.labels.tool.setText(self.dictTool["title"])
        self.labels.citation.setText(self.dictTool["citation"])
        self.labels.description.setText(self.dictTool["description"])
        self.labels.task.setVisible(False)
        self.labels.tool.setVisible(True)
        self.labels.citation.setVisible(True)

    def showTask(self, **kwargs):
        d = {k: v for k, v in kwargs.items() if v is not None}
        d = {k: v for k, v in d.items() if k in self.dictTask.keys()}
        self.dictTask.update(d)
        self.labels.task.setText(self.dictTask["title"])
        self.labels.description.setText(self.dictTask["description"])
        self.labels.tool.setVisible(False)
        self.labels.citation.setVisible(False)
        self.labels.task.setVisible(True)

    def updateLabelWidth(self):
        width = max(self.labels.minimumWidth(), self.labels.sizeHint().width())
        self.labels.setMinimumWidth(width)

    @property
    def logoTool(self):
        return self._logoTool

    @logoTool.setter
    def logoTool(self, logo):
        self.widgetLogoTool.setPixmap(logo)
        self._logoTool = logo

    @property
    def logoProject(self):
        return self.logoProject.logo

    @logoProject.setter
    def logoProject(self, logo):
        self.widgetLogoProject.logo = logo


class HeaderOld(QtWidgets.QFrame):
    """
    The Taxotools toolbar, with space for a title, description,
    citations and two logos.
    """

    def __init__(self):
        """ """
        super().__init__()

        self._title = None
        self._description = None
        self._citation = None
        self._logoTool = None

        self.logoSize = 64

        self.draw()

    def draw(self):
        """ """
        self.setStyleSheet(
            """
            HeaderOld {
                background: palette(Light);
                border-top: 2px solid palette(Mid);
                border-bottom: 1px solid palette(Dark);
                }
            """
        )
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Maximum
        )

        self.labelDescription = QtWidgets.QLabel("DESCRIPTION")
        self.labelDescription.setAlignment(QtCore.Qt.AlignBottom)
        self.labelDescription.setStyleSheet(
            """
            color: palette(Text);
            font-size: 12px;
            font-weight: bold;
            letter-spacing: 1px;
            """
        )

        self.labelCitation = QtWidgets.QLabel("CITATION")
        self.labelCitation.setAlignment(QtCore.Qt.AlignTop)
        self.labelCitation.setStyleSheet(
            """
            color: palette(Shadow);
            font-size: 12px;
            """
        )

        labels = QtWidgets.QVBoxLayout()
        labels.addStretch(1)
        labels.addWidget(self.labelDescription)
        labels.addWidget(self.labelCitation)
        labels.addStretch(1)
        labels.addSpacing(4)
        labels.setSpacing(4)

        self.labelLogoTool = QtWidgets.QLabel()
        self.labelLogoTool.setAlignment(QtCore.Qt.AlignCenter)

        self.labelLogoProject = ScalingImage()
        layoutLogoProject = QtWidgets.QHBoxLayout()
        layoutLogoProject.addWidget(self.labelLogoProject)
        layoutLogoProject.setContentsMargins(2, 4, 2, 4)

        self.toolbar = QtWidgets.QToolBar()
        self.toolbar.setIconSize(QtCore.QSize(32, 32))
        self.toolbar.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum
        )
        self.toolbar.setToolButtonStyle(
            QtCore.Qt.ToolButtonStyle.ToolButtonTextUnderIcon
        )
        self.toolbar.setStyleSheet(
            """
            QToolBar {
                spacing: 2px;
                }
            QToolButton {
                color: palette(ButtonText);
                background: transparent;
                border: 2px solid transparent;
                border-radius: 3px;
                font-size: 14px;
                min-width: 50px;
                min-height: 60px;
                padding: 6px 0px 0px 0px;
                margin: 4px 0px 4px 0px;
                }
            QToolButton:hover {
                background: palette(Window);
                border: 2px solid transparent;
                }
            QToolButton:pressed {
                background: palette(Midlight);
                border: 2px solid palette(Mid);
                border-radius: 3px;
                }
            QToolButton[popupMode="2"]:pressed {
                padding-bottom: 5px;
                border: 1px solid palette(Dark);
                margin: 5px 1px 0px 1px;
                border-bottom-right-radius: 0px;
                border-bottom-left-radius: 0px;
                }
            QToolButton::menu-indicator {
                image: none;
                width: 30px;
                border-bottom: 2px solid palette(Mid);
                subcontrol-origin: padding;
                subcontrol-position: bottom;
                }
            QToolButton::menu-indicator:disabled {
                border-bottom: 2px solid palette(Midlight);
                }
            QToolButton::menu-indicator:pressed {
                border-bottom: 0px;
                }
            """
        )

        layout = QtWidgets.QHBoxLayout()
        layout.addSpacing(8)
        layout.addWidget(self.labelLogoTool)
        layout.addSpacing(6)
        layout.addWidget(VLineSeparator())
        layout.addSpacing(12)
        layout.addLayout(labels, 0)
        layout.addSpacing(12)
        layout.addWidget(VLineSeparator())
        layout.addSpacing(8)
        layout.addWidget(self.toolbar, 0)
        layout.addStretch(1)
        layout.addWidget(VLineSeparator())
        layout.addLayout(layoutLogoProject, 0)
        # layout.addWidget(self.labelLogoProject)
        layout.addSpacing(2)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        self.labelDescription.setText(description)
        self._description = description

    @property
    def citation(self):
        return self._citation

    @citation.setter
    def citation(self, citation):
        self.labelCitation.setText(citation)
        self._citation = citation

    @property
    def logoTool(self):
        return self._logoTool

    @logoTool.setter
    def logoTool(self, logo):
        self.labelLogoTool.setPixmap(logo)
        self._logoTool = logo

    @property
    def logoProject(self):
        return self.labelLogoProject.logo

    @logoProject.setter
    def logoProject(self, logo):
        self.labelLogoProject.logo = logo


class Subheader(QtWidgets.QFrame):
    """A simple styled frame"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Maximum
        )
        self.setStyleSheet(
            """
            QFrame {
                background-color: palette(Midlight);
                border-style: solid;
                border-width: 1px 0px 1px 0px;
                border-color: palette(Mid);
                }
            """
        )


class NavigationButton(PushButton):
    def __init__(self, text, *args, **kwargs):
        super().__init__(text, *args, **kwargs)
        self._default_text = text

    def setText(self, text=None):
        if text is None:
            return super().setText(self._default_text)
        return super().setText(text)


def _mode_method(mode):
    """
    Class method decorator that populates `_mode_methods` with the
    given mode, pointing to the decorated function, for use by setMode().
    """

    class ModeMethod:
        def __init__(self, method):
            self.method = method

        def __call__(self, *args, **kwargs):
            self.method(*args, **kwargs)

        def __set_name__(self, owner, name):
            owner._mode_methods[mode] = self.method
            setattr(owner, name, self.method)

    return ModeMethod


class NavigationFooter(QtWidgets.QFrame):
    """A styled footer with navigation buttons"""

    _mode_methods = {}
    _buttons = []

    class Mode(enum.Enum):
        First = enum.auto()
        Middle = enum.auto()
        Final = enum.auto()
        Wait = enum.auto()
        Error = enum.auto()
        Frozen = enum.auto()

    class ButtonMode(enum.Enum):
        Enabled = enum.auto()
        Disabled = enum.auto()
        Hidden = enum.auto()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setStyleSheet(
            """
            QFrame {
                border-style: solid;
                border-width: 1px 0px 0px 0px;
                border-color: Palette(Mid);
                }
            """
        )

        self.back = self.addButton("< &Back")
        self.next = self.addButton("&Next >")
        self.exit = self.addButton("E&xit")
        self.cancel = self.addButton("&Cancel")
        self.new = self.addButton("&New")

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self.cancel)
        layout.addWidget(self.new)
        layout.addStretch(1)
        layout.addWidget(self.back)
        layout.addWidget(self.next)
        layout.addWidget(self.exit)
        layout.setSpacing(8)
        layout.setContentsMargins(16, 16, 16, 16)
        self.setLayout(layout)

    def addButton(self, text):
        button = NavigationButton(text)
        self._buttons.append(button)
        return button

    def setButtonActions(self, dictionary):
        """Bind functions to buttons"""
        for name in dictionary:
            button = getattr(self, name)
            button.clicked.connect(dictionary[name])

    def setMode(self, mode: Mode, backwards=False):
        """Calls the function corresponding to given mode"""
        if mode in self._mode_methods:
            self._mode_methods[mode](self, backwards)
            for button in self._buttons:
                button.setText()
        else:
            raise ValueError(f"Mode {mode} has no matching method.")

    def setButtonMode(self, button, mode):
        button.setVisible(mode != self.ButtonMode.Hidden)
        button.setEnabled(mode == self.ButtonMode.Enabled)

    @_mode_method(Mode.First)
    def setModeFirst(self, backwards=False):
        self.setButtonMode(self.back, self.ButtonMode.Disabled)
        self.setButtonMode(self.next, self.ButtonMode.Enabled)
        self.setButtonMode(self.exit, self.ButtonMode.Hidden)
        self.setButtonMode(self.cancel, self.ButtonMode.Disabled)
        self.setButtonMode(self.new, self.ButtonMode.Hidden)
        self.next.setFocus()

    @_mode_method(Mode.Middle)
    def setModeMiddle(self, backwards=False):
        self.setButtonMode(self.back, self.ButtonMode.Enabled)
        self.setButtonMode(self.next, self.ButtonMode.Enabled)
        self.setButtonMode(self.exit, self.ButtonMode.Hidden)
        self.setButtonMode(self.cancel, self.ButtonMode.Disabled)
        self.setButtonMode(self.new, self.ButtonMode.Hidden)
        if backwards:
            self.back.setFocus()
        else:
            self.next.setFocus()

    @_mode_method(Mode.Final)
    def setModeFinal(self, backwards=False):
        self.setButtonMode(self.back, self.ButtonMode.Enabled)
        self.setButtonMode(self.next, self.ButtonMode.Hidden)
        self.setButtonMode(self.exit, self.ButtonMode.Enabled)
        self.setButtonMode(self.cancel, self.ButtonMode.Hidden)
        self.setButtonMode(self.new, self.ButtonMode.Enabled)
        self.exit.setFocus()

    @_mode_method(Mode.Wait)
    def setModeWait(self, backwards=False):
        self.setButtonMode(self.back, self.ButtonMode.Disabled)
        self.setButtonMode(self.next, self.ButtonMode.Disabled)
        self.setButtonMode(self.exit, self.ButtonMode.Hidden)
        self.setButtonMode(self.cancel, self.ButtonMode.Enabled)
        self.setButtonMode(self.new, self.ButtonMode.Hidden)
        self.cancel.setFocus()

    @_mode_method(Mode.Error)
    def setModeError(self, backwards=False):
        self.setButtonMode(self.back, self.ButtonMode.Enabled)
        self.setButtonMode(self.next, self.ButtonMode.Disabled)
        self.setButtonMode(self.exit, self.ButtonMode.Hidden)
        self.setButtonMode(self.cancel, self.ButtonMode.Disabled)
        self.setButtonMode(self.new, self.ButtonMode.Hidden)
        self.back.setFocus()

    @_mode_method(Mode.Frozen)
    def setModeFrozen(self, backwards=False):
        self.setButtonMode(self.back, self.ButtonMode.Disabled)
        self.setButtonMode(self.next, self.ButtonMode.Disabled)
        self.setButtonMode(self.exit, self.ButtonMode.Hidden)
        self.setButtonMode(self.cancel, self.ButtonMode.Disabled)
        self.setButtonMode(self.new, self.ButtonMode.Hidden)


class Panel(QtWidgets.QWidget):
    """
    A stylized panel with title, footer and body.
    Set `self.title`, `self.footer` and `self.flag` with text.
    Use `self.body.addWidget()`` to populate the pane.
    """

    def __init__(self, parent):
        """Initialize internal vars"""
        super().__init__(parent=parent)
        self._title = None
        self._foot = None
        self._flag = None
        self._flagTip = None

        # if not hasattr(parent, '_pane_foot_height'):
        #     parent._pane_foot_height = None
        self.draw()

    def draw(self):
        """ """
        self.labelTitle = QtWidgets.QLabel("TITLE GO HERE")
        self.labelTitle.setIndent(4)
        self.labelTitle.setMargin(2)
        self.labelTitle.setStyleSheet(
            """
            QLabel {
                font-size: 14px;
                font-weight: bold;
                letter-spacing: 1px;
                color: palette(Light);
                background: palette(Shadow);
                border-right: 1px solid palette(Dark);
                border-bottom: 2px solid palette(Shadow);
                border-bottom-left-radius: 0px;
                border-top-right-radius: 1px;
                padding-top: 2px;
                }
            QLabel:disabled {
                background: palette(Mid);
                border-right: 1px solid palette(Mid);
                border-bottom: 2px solid palette(Mid);
                }
            """
        )

        self.labelFlag = QtWidgets.QLabel("")
        self.labelFlag.setVisible(False)
        self.labelFlag.setIndent(4)
        self.labelFlag.setMargin(2)
        self.labelFlag.setStyleSheet(
            """
            QLabel {
                font-size: 12px;
                font-weight: bold;
                letter-spacing: 1px;
                color: palette(Light);
                background: palette(Mid);
                border-right: 1px solid palette(Midlight);
                border-bottom: 2px solid palette(Midlight);
                border-bottom-left-radius: 1px;
                border-top-right-radius: 1px;
                padding-top: 1px;
                }
            QLabel:disabled {
                background: palette(Midlight);
                border-right: 1px solid palette(Light);
                border-bottom: 2px solid palette(Light);
                }
            """
        )

        # To be filled by user
        self.body = QtWidgets.QVBoxLayout()

        self.labelFoot = QtWidgets.QLabel("FOOTER")
        self.labelFoot.setAlignment(QtCore.Qt.AlignCenter)
        self.labelFoot.setStyleSheet(
            """
            QLabel {
                color: palette(Shadow);
                background: palette(Window);
                border: 1px solid palette(Mid);
                padding: 5px 10px 5px 10px;
                }
            QLabel:disabled {
                color: palette(Mid);
                background: palette(Window);
                border: 1px solid palette(Mid);
                }
            """
        )

        layoutTop = QtWidgets.QHBoxLayout()
        layoutTop.addWidget(self.labelTitle, 1)
        layoutTop.addWidget(self.labelFlag, 0)
        layoutTop.setSpacing(4)

        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(layoutTop, 0)
        layout.addLayout(self.body, 1)
        layout.addWidget(self.labelFoot, 0)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, title):
        self.labelTitle.setText(title)
        self._title = title

    @property
    def flag(self):
        return self._flag

    @flag.setter
    def flag(self, flag):
        if flag is not None:
            self.labelFlag.setText(flag)
            self.labelFlag.setVisible(True)
        else:
            self.labelFlag.setVisible(False)
        self._flag = flag

    @property
    def flagTip(self):
        return self._flagTip

    @flagTip.setter
    def flagTip(self, flagTip):
        if flagTip is not None:
            self.labelFlag.setToolTip(flagTip)
        else:
            self.labelFlag.setToolTip("")
        self._flagTip = flagTip

    @property
    def footer(self):
        return self._foot

    @footer.setter
    def footer(self, footer):
        self.labelFoot.setVisible(footer != "")
        self.labelFoot.setText(footer)
        self._foot = footer


class ToolDialog(QtWidgets.QWidget):
    """
    For use as the main window of a tool.
    Handles notification sub-dialogs.
    Asks for verification before closing.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowFlags(QtCore.Qt.Window)
        self.title = "iTaxoTools"

    def closeEvent(self, event):
        """Called on window close"""
        if self.reject():
            self.onReject()
            event.accept()
        else:
            event.ignore()

    def reject(self) -> bool:
        """Return True to close window, False to prevent closing"""
        msgBox = QtWidgets.QMessageBox(self)
        msgBox.setWindowTitle(self.title)
        msgBox.setIcon(QtWidgets.QMessageBox.Question)
        msgBox.setText("Are you sure you want to quit?")
        msgBox.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        msgBox.setDefaultButton(QtWidgets.QMessageBox.Yes)
        confirm = self.msgShow(msgBox)
        return confirm == QtWidgets.QMessageBox.Yes

    def onReject(self):
        """Called when dialog is closed"""

    def msgCloseAll(self):
        """Rejects any open QMessageBoxes"""
        for widget in self.children():
            if widget.__class__ == QtWidgets.QMessageBox:
                widget.reject()

    def msgShow(self, dialog):
        """Exec given QMessageBox after closing all others"""
        self.msgCloseAll()
        return dialog.exec()

    def showException(self, exception):
        """Show exception dialog"""
        msgBox = QtWidgets.QMessageBox(self)
        msgBox.setWindowTitle(self.title)
        msgBox.setIcon(QtWidgets.QMessageBox.Critical)
        msgBox.setText("An exception occured:")
        msgBox.setInformativeText(str(exception))
        msgBox.setDetailedText(traceback.format_exc())
        msgBox.setStandardButtons(QtWidgets.QMessageBox.Ok)
        self.msgShow(msgBox)



# -----------------------------------------------------------------------------
# Commons - Utility classes for iTaxoTools modules
# Copyright (C) 2021-2023  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------


class TypeMeta(type):
    _inheritors = dict()

    def __new__(cls, name, bases, attrs):
        obj = super().__new__(cls, name, bases, attrs)
        return cls._patch_object(obj, name, bases)

    @classmethod
    def _patch_object(cls, obj, name, bases):
        cls._inheritors[obj] = dict()
        obj._parent = None
        for base in bases:
            if issubclass(base, Type):
                cls._inheritors[base][name] = obj
                obj._parent = base
        return obj

    def __dir__(self):
        return super().__dir__() + list(self._inheritors[self].keys())

    def __getattr__(self, attr):
        if attr in self._inheritors[self]:
            return self._inheritors[self][attr]
        raise AttributeError(f"Type {repr(self.__name__)} has no subtype {repr(attr)}")

    def __iter__(self):
        return iter(self._inheritors[self].values())


class Type(metaclass=TypeMeta):
    """All subclasses are added as class attributes"""

    def __repr__(self):
        if self._parent:
            return f'<{".".join(self._get_name_chain())}>'
        return super().__repr__()

    def __eq__(self, other):
        return type(self) is type(other)

    @classmethod
    def _get_name_chain(cls):
        chain = [cls.__name__]
        while cls._parent:
            cls = cls._parent
            chain.append(cls.__name__)
        return reversed(chain[:-1])

    @property
    def type(self):
        return type(self)



# -----------------------------------------------------------------------------
# Commons - Utility classes for iTaxoTools modules
# Copyright (C) 2021  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""Miscellaneous widgets"""

from PySide6 import QtCore, QtGui, QtWidgets


class TextEditLogger(QtWidgets.QPlainTextEdit):
    """Thread-safe log display in a QPlainTextEdit"""

    _appendSignal = QtCore.Signal(object)
    _resetSignal = QtCore.Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setReadOnly(True)
        self._appendSignal.connect(self._appendTextInline)
        self._resetSignal.connect(self.clear)

    @QtCore.Slot(object)
    def _appendTextInline(self, text):
        """Using signals ensures thread safety"""
        self.moveCursor(QtGui.QTextCursor.End)
        self.insertPlainText(text)
        sb = self.verticalScrollBar()
        sb.setValue(sb.maximum())
        self.moveCursor(QtGui.QTextCursor.End)

    def append(self, text):
        """Call this to append text to the widget"""
        self._appendSignal.emit(str(text))

    def reset(self):
        self._resetSignal.emit()


class SearchWidget(QtWidgets.QLineEdit):
    """Embedded line edit with search button"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setPlaceholderText("Search")
        self.setStyleSheet(
            """
            SearchWidget {
                padding: 4px;
                padding-left: 4px;
                border: 1px solid Palette(Mid);
                border-radius: 0;
                min-width: 160px;
                }
            SearchWidget:focus {
                border: 1px solid Palette(Highlight);
                }
            """
        )

    def setSearchAction(self, action):
        """Bind a QAction to the widget"""
        self.returnPressed.connect(action.trigger)
        self.addAction(action, QtWidgets.QLineEdit.TrailingPosition)

    def focusInEvent(self, event):
        super().focusInEvent(event)
        QtCore.QTimer.singleShot(0, self.selectAll)



# -----------------------------------------------------------------------------
# Commons - Utility classes for iTaxoTools modules
# Copyright (C) 2021  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""IO redirection and file-likes"""

import io
import os
import sys
from contextlib import contextmanager
from pathlib import Path


@contextmanager
def _redirect(module=sys, stream="stdout", dest=None):
    """Redirect module stream to file stream"""
    original = getattr(module, stream)
    if original is not None:
        original.flush()
    setattr(module, stream, dest)
    try:
        yield dest
    finally:
        dest.flush()
        setattr(module, stream, original)


@contextmanager
def redirect(module=sys, stream="stdout", dest=None, mode="w"):
    """
    Redirect module's stream according to `dest`:
    - If None: Do nothing
    - If String or Path: Open file and redirect
    - Else: Assume IOWrapper, redirect
    """
    if dest is None:
        yield getattr(module, stream)
    elif isinstance(dest, str) or isinstance(dest, Path):
        with open(dest, mode) as file, _redirect(module, stream, file) as f:
            yield f
    else:
        with _redirect(module, stream, dest) as f:
            yield f


class PipeIO(io.IOBase):
    """File-like object that writes to a pipe connection"""

    # ? There are possibly better ways to do this
    # ? Todo- implement read
    def __init__(self, connection, mode):
        super().__init__()
        self._pid = os.getpid()
        self._cache = ""
        self.connection = connection
        self.buffer = ""
        if not (mode == "r" or mode == "w"):
            raise ValueError("Invalid mode: '{}'".format(str(mode)))
        self.mode = mode

    @property
    def cache(self):
        """Fork-safe, discard cache, from multiprocessing doc"""
        pid = os.getpid()
        if pid != self._pid:
            self._pid = pid
            self._cache = ""
        return self._cache

    @cache.setter
    def cache(self, value):
        self._cache = value

    def close(self):
        self.flush()
        self.connection.close()
        self.closed = True

    def fileno(self):
        return self.connection.fileno()

    def readable(self):
        return self.mode == "r"

    def read(self, size=-1):
        if not self.readable():
            raise io.UnsupportedOperation("not readable")
        if size < 0:
            if self.cache == "":
                self.cache = self.connection.recv()
            result = self.cache
            self.cache = ""
            return result
        while len(self.cache) < size:
            self.cache += self.connection.recv()
        result = self.cache[:size]
        self.cache = self.cache[size:]
        return result

    # To do if required:
    # def readline(size=-1):
    #     pass
    # def readlines(hint=-1):
    #     pass

    def writable(self):
        return self.mode == "w"

    def write(self, text):
        if not self.writable():
            raise io.UnsupportedOperation("not writable")
        # no buff, just send
        self.connection.send(text)
        # temp = self.buffer + text
        # self.buffer = ''
        # for line in temp.splitlines(True):
        #     if line[-1] == '\n':
        #         self.connection.send(line)
        #     else:
        #         self.buffer += line

    def writelines(self, lines):
        for line in lines:
            self.connection.send(line + "\n")

    def flush(self):
        if self.buffer != "":
            self.connection.send(self.buffer)
        self.buffer = ""


class TextEditLoggerIO(io.IOBase):
    """File-like object that writes to TextEditLogger"""

    def __init__(self, widget):
        super().__init__()
        self.widget = widget

    def close(self):
        pass

    def flush(self):
        pass

    def readable(self):
        return False

    def writeable(self):
        return True

    def write(self, text):
        self.widget.append(text)

    def writeline(self, line):
        self.write(line + "\n")

    def writelines(self, lines):
        for line in lines:
            self.writeline(line)



# -----------------------------------------------------------------------------
# Commons - Utility classes for iTaxoTools modules
# Copyright (C) 2021-2023  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

from __future__ import annotations

from PySide6 import QtCore

from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from types import UnionType
from typing import Callable, Iterator, Optional, TypeVar, Union, get_origin

from .utility import AttrDict


class Instance:
    """Pass as type or instance to instantiate a property by default"""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class _Instance:
    """Internal instructions for default class instatiation"""

    def __init__(self, type, *args, **kwargs):
        self.type = type
        self.args = args
        self.kwargs = kwargs


class Property:
    key_ref = "properties"
    key_list = "_property_list"
    key_tags = "_property_tags"

    def __init__(self, type=object, default=None, tag=None):
        if isinstance(type, UnionType):
            type = object
        self.type = type
        self.default = default
        self.tag = tag

    @staticmethod
    def key_value(key):
        return f"_property_{key}_value"

    @staticmethod
    def key_notify(key):
        return f"_property_{key}_notify"

    @staticmethod
    def key_getter(key):
        return f"_property_{key}_getter"

    @staticmethod
    def key_setter(key):
        return f"_property_{key}_setter"

    @staticmethod
    def key_default(key):
        return f"_property_{key}_default"


class PropertyRef:
    def __init__(self, parent, key):
        self._parent = parent
        self._key = key

    @property
    def notify(self) -> Callable[[], None]:
        return getattr(self._parent, Property.key_notify(self._key))

    @property
    def get(self) -> Callable[[], object]:
        return getattr(self._parent, Property.key_getter(self._key))

    @property
    def set(self) -> Callable[[object], None]:
        return getattr(self._parent, Property.key_setter(self._key))

    @property
    def default(self) -> object:
        default = getattr(self._parent, Property.key_default(self._key))
        if isinstance(default, _Instance):
            return default.type(*default.args, **default.kwargs)
        return default

    @property
    def key(self) -> str:
        return self._key

    @property
    def value(self) -> object:
        return self.get()

    @property
    def tag(self) -> object:
        tags = getattr(self._parent, Property.key_tags)
        return tags[self._key]

    @value.setter
    def value(self, value):
        return self.set(value)

    def update(self) -> Callable[[], object]:
        self.notify.emit(self.get())


class PropertiesRef:
    def __init__(self, parent):
        self._parent = parent

    def __getattr__(self, attr):
        if attr in self._list():
            return PropertyRef(self._parent, attr)
        raise AttributeError(
            f"{repr(type(self._parent).__name__)} has no property: {repr(attr)}"
        )

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __dir__(self):
        return super().__dir__() + self._list()

    def _list(self):
        return getattr(self._parent, Property.key_list)

    def __iter__(self) -> Iterator[PropertyRef]:
        return (self[key] for key in self._list())

    def __contains__(self, key):
        return key in self._list()


class PropertyMeta(type(QtCore.QObject)):
    def __new__(cls, name, bases, attrs):
        properties = {
            key: attrs[key] for key in attrs if isinstance(attrs[key], Property)
        }
        cls._init_list(bases, attrs)
        for key, prop in properties.items():
            cls._register_property(attrs, key, prop)
        cls._add_ref(attrs)
        obj = super().__new__(cls, name, bases, attrs)
        return obj

    def _init_list(bases, attrs):
        key_list = Property.key_list
        lists = [getattr(base, key_list, []) for base in bases]
        attrs[key_list] = sum(lists, [])

        key_tags = Property.key_tags
        attrs[key_tags] = defaultdict(lambda: None)

    def _register_property(attrs: dict[str, object], key: str, prop: Property):
        key_value = Property.key_value(key)
        key_notify = Property.key_notify(key)
        key_getter = Property.key_getter(key)
        key_setter = Property.key_setter(key)
        key_default = Property.key_default(key)
        key_list = Property.key_list
        key_tags = Property.key_tags

        if prop.tag is not None:
            attrs[key_tags][key] = prop.tag

        prop_type = get_origin(prop.type) or prop.type
        if isinstance(prop_type, TypeVar) and prop_type.__bound__:
            prop_type = prop_type.__bound__

        notify = QtCore.Signal(prop_type)

        def getter(self):
            return getattr(self, key_value, None)

        def setter(self, value):
            old = getattr(self, key_value, None)
            setattr(self, key_value, value)
            if old != value:
                getattr(self, key_notify).emit(value)

        default = prop.default
        if default == Instance:
            default = _Instance(prop_type)
        if isinstance(default, Instance):
            default = _Instance(prop_type, *default.args, **default.kwargs)

        attrs[key_list].append(key)

        attrs[key_notify] = notify
        attrs[key_getter] = getter
        attrs[key_setter] = setter
        attrs[key_default] = default

        attrs[key] = QtCore.Property(
            type=prop_type,
            fget=getter,
            fset=setter,
            notify=notify,
        )

    def _add_ref(attrs):
        key_ref = Property.key_ref

        def getref(self):
            return PropertiesRef(self)

        attrs[key_ref] = property(getref)


class PropertyObject(QtCore.QObject, metaclass=PropertyMeta):
    properties: PropertiesRef

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._set_property_defaults()

    def _set_property_defaults(self):
        for property in self.properties:
            property.set(property.default)

    def as_dict(self):
        return AttrDict({property.key: property.value for property in self.properties})


class EnumObjectMeta(PropertyMeta):
    def __new__(cls, name, bases, attrs):
        enum = attrs.get("enum", None)
        if not enum:
            return super().__new__(cls, name, bases, attrs)

        get_key = attrs.get("get_key", lambda x: x.key)
        get_type = attrs.get("get_type", lambda x: x.type)
        get_default = attrs.get("get_default", lambda x: x.default)
        for field in enum:
            attrs[get_key(field)] = Property(get_type(field), get_default(field))
        return super().__new__(cls, name, bases, attrs)


class EnumObject(PropertyObject, metaclass=EnumObjectMeta):
    enum: Enum

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "enum" in dir(self):
            self.reset()

    def reset(self):
        self._set_property_defaults()


BindingHash = tuple[QtCore.SignalInstance, Callable]


@dataclass(frozen=True)
class Binding:
    source: Union[PropertyRef, QtCore.SignalInstance]
    destination: Union[PropertyRef, Callable]
    signal: QtCore.SignalInstance
    slot: Callable
    bound_slot: Callable
    update_slot: Callable

    def __post_init__(self):
        self.update()

    @staticmethod
    def _get_proxied_slot(slot, proxy):
        if not proxy:
            return slot

        def proxy_slot(value):
            slot(proxy(value))

        return proxy_slot

    @staticmethod
    def _get_conditional_slot(slot, condition):
        if not condition:
            return slot

        def conditional_slot(value):
            if condition(value):
                slot(value)

        return conditional_slot

    @classmethod
    def new(
        cls,
        source: Union[PropertyRef, QtCore.SignalInstance],
        destination: Union[PropertyRef, Callable],
        proxy: Optional[Callable] = None,
        condition: Optional[Callable] = None,
    ) -> Binding:
        if isinstance(source, PropertyRef):
            signal = source.notify
            update_slot = source.update
        else:
            signal = source
            update_slot = None

        if isinstance(destination, PropertyRef):
            slot = destination.set
        else:
            slot = destination

        bound_slot = slot
        bound_slot = cls._get_proxied_slot(bound_slot, proxy)
        bound_slot = cls._get_conditional_slot(bound_slot, condition)

        signal.connect(bound_slot)

        return cls(
            source=source,
            destination=destination,
            signal=signal,
            slot=slot,
            bound_slot=bound_slot,
            update_slot=update_slot,
        )

    def unbind(self):
        self.signal.disconnect(self.bound_slot)

    def update(self):
        if self.update_slot is not None:
            self.update_slot()


class Binder(dict[BindingHash, Binding]):
    def bind(
        self,
        source: Union[PropertyRef, QtCore.SignalInstance],
        destination: Union[PropertyRef, Callable],
        proxy: Optional[Callable] = None,
        condition: Optional[Callable] = None,
    ) -> None:
        binding = Binding.new(source, destination, proxy, condition)
        self[self._hash(source, destination)] = binding

    def unbind(
        self,
        source: Union[PropertyRef, QtCore.SignalInstance],
        destination: Union[PropertyRef, Callable],
    ) -> None:
        hash = self._hash(source, destination)
        self[hash].unbind()
        del self[hash]

    def unbind_all(self):
        for binding in self.values():
            binding.unbind()
        self.clear()

    def update(self):
        for binding in self.values():
            binding.update()

    def _hash(
        self,
        source: Union[PropertyRef, QtCore.SignalInstance],
        destination: Union[PropertyRef, Callable],
    ) -> BindingHash:
        if isinstance(source, PropertyRef):
            signal = source.notify
        else:
            signal = source

        if isinstance(destination, PropertyRef):
            slot = destination.set
        else:
            slot = destination

        return (signal, slot)



# -----------------------------------------------------------------------------
# Commons - Utility classes for iTaxoTools modules
# Copyright (C) 2021  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""Extensions for QStateMachines"""


from PySide6 import QtStateMachine


class TaggedEvent(QtStateMachine.QStateMachine.SignalEvent):
    """Wraps a SignalEvent meant for a TaggedTransition."""

    def __init__(self, event):
        if self.isValidCast(event):
            super().__init__(event.sender(), event.signalIndex(), [])
            self.tag = event.arguments()[0]
            self.args = event.arguments()[1]
            self.kwargs = event.arguments()[2]
        else:
            super().__init__(None, -1, [])
            self.type = event.type
            self.tag = None
            self.args = []
            self.kwargs = {}

    @staticmethod
    def isValidCast(event):
        return (
            isinstance(event, QtStateMachine.QStateMachine.SignalEvent)
            and len(event.arguments()) == 3
            and isinstance(event.arguments()[1], list)
            and isinstance(event.arguments()[2], dict)
        )


class TaggedTransition(QtStateMachine.QSignalTransition):
    """Only catches TaggedEvent with matching tags."""

    def __init__(
        self,
        signal,  # created with: Signal(object, list, dict)
        tag,
        filter=lambda e: True,
        effect=lambda e: None,
    ):
        super().__init__(signal)
        self.tag = tag
        self.filter = filter
        self.effect = effect

    def eventTest(self, event):
        taggedEvent = TaggedEvent(event)
        if self.tag == taggedEvent.tag:
            return self.filter(event)
        return False

    def onTransition(self, event):
        taggedEvent = TaggedEvent(event)
        self.effect(taggedEvent)
        super().onTransition(event)



# -----------------------------------------------------------------------------
# Param - Parameter primitives
# Copyright (C) 2021  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------


"""
Parameter primitives to hold values and metadata.
First define the model as a root Group with children.
Then access the Field values using Group attributes.

Usage example:

    from itaxotools.common.param import Field, Group

    param = Group(
        key='root',
        label='General',
        children=[
            Field('alpha', default='lorem'),
            Field('beta', value=42, doc='The answer'),
            Group(
                key='general',
                children=[
                    Field('gamma', list=[1, 2, 3], value=2),
                    Field('delta', type=int, value=1),
                    Field('epsilon', range=(1, 10), value=1),
                ])
        ])

    param.alpha = 'ipsum'
    print(param.beta.doc, param.beta.value)
    if param.beta == 42:
        print('This is True')
    try:
        param.general.gamma = 5
    except ValueError as e:
        print('Exception:', str(e))
    try:
        param.general.delta = '5'
    except TypeError as e:
        print('Exception:', str(e))
    data = param.dumps()
    data['general']['epsilon'] = 10
    param.loads(data)
    print(repr(param.general))

"""

from copy import copy


class Field:
    """
    Contains the Field value and other metadata.
    New values are cheked against type, list and range (if defined).
    Using comparison operators on the Field compare against its value.
    """

    def __init__(self, key, **kwargs):
        """
        The key must be a valid Python variable name.
        All keyword arguments become attributes.
        """
        self.key = None
        self.doc = None
        self.default = None
        self.list = {}
        self.range = (None, None)
        self._label = None
        self._value = None
        self._parent = None
        self._type = type(None)

        if isinstance(key, str) and key.isidentifier():
            self.key = key
        else:
            raise TypeError(f"key {repr(key)} is not a valid string identifier")

        # Assign these in order
        if "label" in kwargs:
            self._label = kwargs.pop("label")
        if "type" in kwargs:
            self.type = kwargs.pop("type")
        if "list" in kwargs:
            self.list = kwargs.pop("list")
        if "range" in kwargs:
            self.range = kwargs.pop("range")
        if "default" in kwargs:
            self.default = kwargs.pop("default")
            self.value = self.default
        if "value" in kwargs:
            self.value = kwargs.pop("value")

        for attr in kwargs:
            setattr(self, attr, kwargs[attr])

    def __str__(self):
        return f"{self.__class__.__name__}('{self.label}': {self.value})"

    def __repr__(self):
        parent = (
            None
            if self._parent is None
            else (
                f"{self._parent.__class__.__name__}"
                f"({repr(self._parent.key)} at {hex(id(self._parent))})"
            )
        )
        return (
            f"{self.__class__.__name__}("
            f"key={repr(self.key)}, "
            f"value={repr(self.value)}, "
            f"type={repr(self.type)}, "
            f"label={repr(self._label)}, "
            f"doc={repr(self.doc)}, "
            f"default={repr(self.default)}, "
            f"list={repr(self.list)}, "
            f"range={repr(self.range)}, "
            f"parent={parent}"
            f") at {hex(id(self))}"
        )

    def __eq__(self, other):
        return self.value == other

    def __ne__(self, other):
        return self.value != other

    def __lt__(self, other):
        return self.value < other

    def __gt__(self, other):
        return self.value > other

    def __le__(self, other):
        return self.value <= other

    def __ge__(self, other):
        return self.value >= other

    @property
    def label(self):
        return self._label if self._label is not None else self.key

    def _get_type(self):
        return self._type

    def _set_type(self, x):
        if not isinstance(x, type):
            raise TypeError(f"new type {x} is not a type")
        self._type = x

    def _get_value(self):
        return self._value

    def _set_value(self, x):
        """Check new value against type, range and list"""
        if not issubclass(self.type, type(None)):
            if not isinstance(x, self.type):
                raise TypeError(
                    (
                        f"new value {repr(x)} "
                        f"for field {repr(self.key)} "
                        f"not of type {self.type.__name__}"
                    )
                )
        if self.list and x not in self.list:
            raise ValueError(
                (f"new value {repr(x)} " f"for field {repr(self.key)} " f"not in list")
            )
        if self.range[0] is not None and x < self.range[0]:
            raise ValueError(
                (
                    f"new value {repr(x)} "
                    f"for field {repr(self.key)} "
                    f"is less than {self.range[0]}"
                )
            )
        if self.range[1] is not None and x > self.range[1]:
            raise ValueError(
                (
                    f"new value {repr(x)} "
                    f"for field {repr(self.key)} "
                    f"is more than {self.range[1]}"
                )
            )
        self._value = x

    def copy(self):
        return copy(self)

    type = property(_get_type, _set_type)
    value = property(_get_value, _set_value)


class Group(object):
    """
    Contains the Group children and other metadata.
    Access the children either by key or attribute.
    Assignment operator on a Field child sets its value.
    Children are added in order.
    """

    def __init__(self, key, **kwargs):
        """
        The key must be a valid Python variable name.
        All keyword arguments become attributes.
        If a children list is provided, they are properly added.
        """
        self.key = None
        self.doc = None
        self._label = None
        self._children = {}
        self._parent = None

        if isinstance(key, str) and key.isidentifier():
            self.key = key
        else:
            raise TypeError(f"key {repr(key)} is not a valid string identifier")

        if "label" in kwargs:
            self._label = kwargs.pop("label")
        if "children" in kwargs:
            for child in kwargs["children"]:
                self.add(child)
            kwargs.pop("children")

        for attr in kwargs:
            setattr(self, attr, kwargs[attr])

    def __str__(self):
        return (
            f"{self.__class__.__name__}('{self.label}': "
            f"{list(self._children.keys())})"
        )

    def __repr__(self):
        children = ", ".join(
            [
                (
                    f"{self._children[x].__class__.__name__}"
                    f"({repr(x)} at {hex(id(self._children[x]))})"
                )
                for x in self._children
            ]
        )
        parent = (
            None
            if self._parent is None
            else (
                f"{self._parent.__class__.__name__}"
                f"({repr(self._parent.key)} at {hex(id(self._parent))})"
            )
        )
        return (
            f"{self.__class__.__name__}("
            f"key={repr(self.key)}, "
            f"label={repr(self._label)}, "
            f"doc={repr(self.doc)}, "
            f"children=[{children}], "
            f"parent={parent}"
            f") at {hex(id(self))}"
        )

    def __getitem__(self, child):
        return self._children[child]

    def __iter__(self):
        return iter(self._children[child] for child in self._children)

    def __delitem__(self, child):
        self._children[child]._parent = None
        del self._children[child]

    def __setitem__(self, key, x):
        if key in self._children and isinstance(self._children[key], Field):
            self._children[key].value = x
        else:
            raise TypeError(f"Group {repr(self.key)} has no Field '{key}'")

    def __getattr__(self, attr):
        try:
            super().__getattribute__("_children")
        except AttributeError:
            return super().__getattribute__(attr)
        else:
            if attr in self._children:
                return self[attr]
            else:
                raise AttributeError(f"Group has no attribute or child {repr(attr)}")

    def __setattr__(self, attr, x):
        if hasattr(self, "_children") and attr in self._children:
            self[attr] = x
        else:
            super().__setattr__(attr, x)

    def __delattr__(self, attr):
        if hasattr(self, "_children") and attr in self._children:
            del self[attr]
        else:
            super().__delattr__(attr)

    def __dir__(self):
        return super().__dir__() + list(self._children.keys())

    @property
    def label(self):
        return self._label if self._label is not None else self.key

    def add(self, x):
        """Add a Field or Group while properly setting key and parent"""
        if not (isinstance(x, Field) or isinstance(x, Group)):
            raise TypeError(f"new child {x} is not a Field or Group")
        if x.key in super().__dir__():
            raise ValueError(f"new child key {repr(x.key)} conflicts with Group dir")
        self._children[x.key] = x
        x._parent = self

    def dumps(self):
        """Export all values as a nested dictionary"""
        d = dict()
        for key, child in self._children.items():
            if isinstance(child, Field):
                d[key] = child.value
            elif isinstance(child, Group):
                d[key] = child.dumps()
            else:
                raise TypeError(f"unexpected child {repr(child)} of type {type(child)}")
        return d

    def loads(self, data):
        """Import values from a nested dictionary"""
        for key, value in data.items():
            if isinstance(self._children[key], Field):
                self[key] = value
            elif isinstance(self._children[key], Group):
                self[key].loads(value)
            else:
                raise KeyError(f"no child for key {repr(key)}")



# PyInstaller entry points for setuptools
# https://pyinstaller.readthedocs.io/en/stable/hooks.html

# When a module is detected by PyInstaller, it will search
# for corresponding hooks and tests in this directory.

from pathlib import Path

import itaxotools


def get_namespace_dirs():
    paths = [Path(path) for path in itaxotools.__path__]
    return [str(path / "__pyinstaller") for path in paths]


def get_hook_dirs():
    return get_namespace_dirs()


def get_PyInstaller_tests():
    return get_namespace_dirs()



# -----------------------------------------------------------------------------
# Commons - Utility classes for iTaxoTools modules
# Copyright (C) 2021  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""Get resource paths for Python3.8+"""


import os
import pathlib
import sys


def _package_path_pyinstaller(package):
    if isinstance(package, str):
        parts = package.split(".")
    elif isinstance(package, type(sys)):
        parts = package.__name__.split(".")
    else:
        return None
    path = pathlib.Path(sys._MEIPASS)
    for part in parts:
        path /= part
    return path


def _package_path_file(package):
    if isinstance(package, str):
        file = sys.modules[package].__file__
    elif isinstance(package, type(sys)):
        file = package.__file__
    else:
        return None
    path = pathlib.Path(os.path.dirname(file))
    return path


def _package_path_importlib(package):
    return importlib.resources.files(package)


try:
    import importlib.resources

    package_path = _package_path_importlib
except (ModuleNotFoundError, ImportError):
    if hasattr(sys, "_MEIPASS"):
        package_path = _package_path_pyinstaller
    else:
        package_path = _package_path_file

_resource_path = package_path(__package__)


def get_common(path):
    return str(_resource_path / path)


def get_local(package, path):
    return str(package_path(package) / path)


def get(*args):
    if len(args) == 1:
        return get_common(*args)
    return get_local(*args)



# -----------------------------------------------------------------------------
# Commons - Utility classes for iTaxoTools modules
# Copyright (C) 2021  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""Simple frames"""


from PySide6 import QtWidgets


class Frame(QtWidgets.QFrame):
    """A slightly darker than the background frame"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFrameStyle(QtWidgets.QFrame.StyledPanel)
        self.setStyleSheet(
            """
            Frame {
                background: rgba(0, 0, 0, 4);
                border: 1px solid rgba(0, 0, 0, 24);
            }"""
        )


class VLineSeparator(QtWidgets.QFrame):
    """Vertical line separator"""

    def __init__(self, width=2):
        super().__init__()
        self.setFixedWidth(width)
        self.setFrameShape(QtWidgets.QFrame.VLine)
        self.setFrameShadow(QtWidgets.QFrame.Plain)
        self.setStyleSheet(
            """
            background: palette(Mid);
            border: none;
            margin: 4px;
            """
        )


class HLineSeparator(QtWidgets.QFrame):
    """Vertical line separator"""

    def __init__(self, height=2):
        super().__init__()
        self.setFixedHeight(height)
        self.setFrameShape(QtWidgets.QFrame.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Plain)
        self.setStyleSheet(
            """
            background: palette(Mid);
            border: none;
            margin: 4px;
            """
        )



# -----------------------------------------------------------------------------
# Param - Parameter model
# Copyright (C) 2021  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------


"""
Parameter model for PySide6
"""

from PySide6.QtCore import QAbstractItemModel, QModelIndex, Qt

from . import Field, Group


class Model(QAbstractItemModel):
    DataRole = Qt.ItemDataRole.UserRole
    columns = ["key", "label", "value", "doc", "list", "range", "type", "default"]

    def __init__(self, group, parent=None):
        super().__init__(parent)
        self.rootItem = group
        self.headers = list(self.columns)

    def index(self, row, column, parent=QModelIndex()):
        if not self.hasIndex(row, column, parent):
            return QModelIndex()
        parentItem = self.rootItem
        if parent.isValid():
            parentItem = parent.internalPointer()
        try:
            keys = list(parentItem._children.keys())
            key = keys[row]
            childItem = parentItem[key]
        except KeyError:
            return QModelIndex()
        return self.createIndex(row, column, childItem)

    def parent(self, index):
        if not index.isValid():
            return QModelIndex()
        childItem = index.internalPointer()
        parentItem = childItem._parent
        if parentItem == self.rootItem or parentItem is None:
            return QModelIndex()
        grandParentItem = parentItem._parent
        if grandParentItem is None:
            return QModelIndex()

        keys = list(grandParentItem._children.keys())
        row = keys.index(parentItem.key)
        return self.createIndex(row, 0, parentItem)

    def rowCount(self, parent):
        if parent.column() > 0:
            return 0
        parentItem = self.rootItem
        if parent.isValid():
            parentItem = parent.internalPointer()
        if isinstance(parentItem, Group):
            return len(parentItem._children)
        return 0

    def columnCount(self, parent):
        return len(self.columns)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        item = index.internalPointer()
        if role == Qt.ItemDataRole.DisplayRole:
            try:
                attr = self.columns[index.column()]
                return str(getattr(item, attr))
            except AttributeError:
                return None
        elif (
            role == Qt.ItemDataRole.ToolTipRole or role == Qt.ItemDataRole.WhatsThisRole
        ):
            return item.doc
        elif role == self.DataRole:
            return item
        return None

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        try:
            if index.isValid() and role == Qt.ItemDataRole.EditRole:
                item = index.internalPointer()
                item.value = item.type(value)
                roles = [Qt.ItemDataRole.DisplayRole, self.DataRole]
                self.dataChanged.emit(index, index, roles)
                return True
            return False
        except (ValueError, TypeError) as e:
            self.setDataError = str(e)
            return False

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        item = index.internalPointer()
        flag = Qt.ItemFlag.NoItemFlags
        if isinstance(item, Field) and self.columns[index.column()] == "value":
            flag = Qt.ItemFlag.ItemIsEditable
        return super().flags(index) | flag

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if (
            orientation == Qt.Orientation.Horizontal
            and role == Qt.ItemDataRole.DisplayRole
        ):
            return self.headers[section]
        return None

    def setHeaderData(
        self, section, orientation, value, role=Qt.ItemDataRole.DisplayRole
    ):
        if (
            orientation == Qt.Orientation.Horizontal
            and role == Qt.ItemDataRole.DisplayRole
        ):
            self.headers[section] = value
            self.headerDataChanged.emit(orientation, section, section)
            return True
        return False

    def paramIndex(self, param):
        if param._parent is None:
            return QModelIndex()
        keys = list(param._parent._children.keys())
        row = keys.index(param.key)
        return self.createIndex(row, 0, param)

    def resetParams(self, index=QModelIndex()):
        rows = self.rowCount(index)
        if rows == 0:
            return
        for row in range(0, rows):
            childIndex = self.index(row, 0, index)
            childItem = childIndex.internalPointer()
            if isinstance(childItem, Field):
                childItem.value = childItem.default
            self.resetParams(childIndex)
        topLeft = self.index(0, 0, index)
        bottomRight = self.index(rows - 1, 0, index)
        roles = [Qt.ItemDataRole.DisplayRole, self.DataRole]
        self.dataChanged.emit(topLeft, bottomRight, roles)



# -----------------------------------------------------------------------------
# Param - Parameter view
# Copyright (C) 2021  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------


"""
Parameter view for PySide6
"""

from PySide6.QtCore import QModelIndex, QRect, QSize, Qt, Signal
from PySide6.QtGui import QDoubleValidator, QIntValidator, QValidator
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QFrame,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLayout,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QScrollArea,
    QVBoxLayout,
    QWidget,
)

from . import Field


class FieldLayout(QLayout):
    """Last widget added has the same width for all instances"""

    def __init__(self, label, widget, parent=None):
        super().__init__(parent)
        self.itemList = []
        self.www = 120
        self.gap = 6
        self.addWidget(label)
        self.addWidget(widget)

    def addItem(self, item):
        self.itemList.append(item)

    def itemAt(self, index):
        try:
            return self.itemList[index]
        except IndexError:
            return None

    def takeAt(self, index):
        return self.itemList.pop(index)

    def count(self):
        return len(self.itemList)

    def sizeHint(self):
        w, h = 0, 0
        for item in self.itemList:
            h = max(h, item.sizeHint().height())
        w += self.itemList[0].sizeHint().width()
        w += self.gap + self.www + 10
        return QSize(w, h)

    def minimumSize(self):
        if self.itemList:
            return self.itemList[-1].sizeHint()
        else:
            return QSize(0, 0)

    def setGeometry(self, rect):
        super().setGeometry(rect)
        w = rect.width()
        if w >= self.www:
            div = rect.right() - self.www
            rl = QRect(rect)
            rl.setRight(div - self.gap)
            self.itemList[0].setGeometry(rl)
            rw = QRect(rect)
            rw.setLeft(div)
            self.itemList[-1].setGeometry(rw)


class FieldWidget(QFrame):
    # Emit (_index, value) when user changes the field value
    dataChanged = Signal(object, object)

    def __init__(self, index, field, view, parent=None):
        super().__init__(parent)
        view.dataChanged.connect(self.onModelDataChange)
        self.dataChanged.connect(view.onWidgetDataChange)
        self._index = index
        self._field = field
        self._view = view

    def onModelDataChange(self, index):
        """Refresh data if model index was updated"""
        if index == self._index:
            self.refreshData()

    def refreshData(self):
        """Called by view when field value is changed"""
        raise NotImplementedError()


class BoolWidget(FieldWidget):
    dataChanged = Signal(object, object)

    def __init__(self, index, field, view, parent=None):
        super().__init__(index, field, view, parent)
        self.checkbox = QCheckBox()
        self.checkbox.setText(self._field.label)
        self.checkbox.stateChanged.connect(self.onDataChange)
        self.checkbox.setToolTip(self._field.doc)
        self.refreshData()
        layout = QHBoxLayout()
        layout.addWidget(self.checkbox)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def onDataChange(self, state):
        v = self.checkbox.isChecked()
        self.dataChanged.emit(self._index, v)

    def refreshData(self):
        data = self._field.value
        self.checkbox.blockSignals(True)
        self.checkbox.setChecked(data)
        self.checkbox.blockSignals(False)

    def setFocus(self, reason=Qt.OtherFocusReason):
        self.checkbox.setFocus(reason)

    def sizeHint(self):
        return self.checkbox.sizeHint() + QSize(5, 0)


class EmptyOrIntValidator(QIntValidator):
    def validate(self, input, pos):
        if not input:
            return QValidator.Acceptable
        return super().validate(input, pos)


class EmptyOrDoubleValidator(QDoubleValidator):
    def validate(self, input, pos):
        if not input:
            return QValidator.Acceptable
        return super().validate(input, pos)


class EntryWidget(FieldWidget):
    def __init__(self, index, field, view, parent=None):
        super().__init__(index, field, view, parent)
        self.label = QLabel(self._field.label + ": ")
        self.entry = QLineEdit()
        self.entry.editingFinished.connect(self.onDataChange)
        if issubclass(self._field.type, int):
            self.entry.setValidator(EmptyOrIntValidator(self.entry))
        elif issubclass(self._field.type, float):
            self.entry.setValidator(EmptyOrDoubleValidator(self.entry))
        self.label.setToolTip(self._field.doc)
        self.entry.setToolTip(self._field.doc)
        self.refreshData()
        layout = FieldLayout(self.label, self.entry)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def onDataChange(self):
        v = self.entry.text()
        self.dataChanged.emit(self._index, v)

    def refreshData(self):
        self.entry.setText(str(self._field.value))

    def setFocus(self, reason=Qt.OtherFocusReason):
        self.entry.setFocus(reason)


class NoWheelComboBox(QComboBox):
    def wheelEvent(self, event):
        if self.hasFocus:
            event.ignore()
        else:
            self.wheelEvent(event)


class ListWidget(FieldWidget):
    def __init__(self, index, field, view, parent=None):
        super().__init__(index, field, view, parent)
        self.label = QLabel(self._field.label + ": ")
        self.combo = NoWheelComboBox()
        # self.combo.setFocusPolicy(Qt.StrongFocus)
        self.combo.currentIndexChanged.connect(self.onDataChange)
        self.populateCombo()
        self.label.setToolTip(self._field.doc)
        self.combo.setToolTip(self._field.doc)
        layout = FieldLayout(self.label, self.combo)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def populateCombo(self):
        self.combo.blockSignals(True)
        self.combo.clear()
        if isinstance(self._field.list, list):
            for k in self._field.list:
                self.combo.addItem(str(k), k)
        elif isinstance(self._field.list, dict):
            for k, v in self._field.list.items():
                self.combo.addItem(v, k)
        self.refreshData()
        self.combo.blockSignals(False)

    def onDataChange(self):
        if self.combo.count() > 0:
            v = self.combo.currentData()
            self.dataChanged.emit(self._index, v)

    def refreshData(self):
        data = self._field.value
        i = list(self._field.list).index(data)
        self.combo.blockSignals(True)
        self.combo.setCurrentIndex(i)
        self.combo.blockSignals(False)

    def setFocus(self, reason=Qt.OtherFocusReason):
        self.combo.setFocus(reason)


class View(QScrollArea):
    """Widget-based view for param.Model"""

    # Emited for every index when model data change
    dataChanged = Signal(object)

    def __init__(self, model=None, showResetButton=True, parent=None):
        super().__init__(parent)
        self.showResetButton = showResetButton
        self._model = None
        self._rootIndex = QModelIndex()
        self._widgets = dict()
        self._customParamClass = dict()
        self.setWidgetResizable(False)
        if model is not None:
            self.setModel(model)

    def rootIndex(self):
        return self._rootIndex

    def setRootIndex(self, index):
        self._rootIndex = index
        self.draw()

    def model(self):
        return self._model

    def setModel(self, model):
        self._widgets = dict()
        self._model = model
        model.dataChanged.connect(self.onModelDataChange)
        self.setRootIndex(QModelIndex())
        self.draw()

    def draw(self):
        widget = self._populate(self._rootIndex, 0)
        self.setWidget(widget)
        widget.setStyleSheet(
            """
            FieldWidget[depth="1"] {
                margin-left: 1px;
                margin-right: 10px;
                }
            QGroupBox:!flat {
                background-color: rgba(0,0,0,0.02);
                border: 1px solid Palette(Mid);
                margin-top: 28px;
                padding: 0px;
                }
            QGroupBox::title:!flat {
                color: Palette(Text);
                subcontrol-origin: margin;
                subcontrol-position: top left;
                padding: 4px -1px;
                }
            QGroupBox:flat {
                border-top: 1px solid Palette(Mid);
                border-left: 0px;
                border-bottom: 0px;
                margin-top: 24px;
                margin-right: -20px;
                padding-left: 12px;
                padding-bottom: 0px;
                }
            QGroupBox::title:flat {
                color: Palette(Text);
                subcontrol-origin: margin;
                subcontrol-position: top left;
                padding: 4px -1px;
                }
        """
        )
        widget.show()

    def onInvalidValue(self, error):
        """Called when a field value was invalid"""
        QMessageBox.warning(self, "Warning", error)

    def addCustomParamWidget(self, param, _class):
        """The view for `param` will be an instance of `_class`"""
        self._customParamClass[id(param)] = _class
        self.draw()

    def onModelDataChange(self, topLeft, bottomRight, roles):
        """Send signal updates for all indices changed"""
        if bottomRight.parent() != topLeft.parent():
            raise RuntimeError("View: Model dataChanged items have different parents")
        parent = topLeft.parent()
        top = topLeft.row()
        bottom = bottomRight.row()
        for row in range(top, bottom + 1):
            index = self._model.index(row, 0, parent)
            self.dataChanged.emit(index)

    def onWidgetDataChange(self, index, data):
        """Update model"""
        if not self._model.setData(index, data):
            self.onInvalidValue(self._model.setDataError)
            self.sender().refreshData()
            self.sender().setFocus()
            # QTimer.singleShot(0, self.sender().setFocus)

    def _fieldWidget(self, index, field, depth=0):
        """Generate and return an appropriate widget for the field"""
        type_to_widget = {
            bool: BoolWidget,
            float: EntryWidget,
            str: EntryWidget,
            int: EntryWidget,
        }
        if id(field) in self._customParamClass:
            _class = self._customParamClass[id(field)]
            widget = _class(index, field, self)
        elif field.list:
            widget = ListWidget(index, field, self)
        elif field.type in type_to_widget.keys():
            widget = type_to_widget[field.type](index, field, self)
        else:
            widget = EntryWidget(index, field, self)
        widget.setProperty("depth", depth)
        return widget

    def _populate(self, index, depth=0):
        """Recursively populate and return a widget with params"""
        data = self._model.data(index, self._model.DataRole)

        if isinstance(data, Field):
            return self._fieldWidget(index, data, depth)

        if depth == 0:
            widget = QFrame()
            layout = QVBoxLayout()
        # elif depth == 1:
        #     widget = QGroupBox(data.label)
        #     layout = QVBoxLayout()
        else:
            widget = QGroupBox(data.label)
            widget.setFlat(True)
            layout = QVBoxLayout()
            layout.setContentsMargins(0, 6, 0, 3)

        rows = self._model.rowCount(index)
        for row in range(0, rows):
            childIndex = self._model.index(row, 0, index)
            child = self._populate(childIndex, depth + 1)
            layout.addWidget(child)

        if self.showResetButton and depth == 0:
            resetButton = QPushButton("Reset to defaults")
            resetButton.setAutoDefault(False)
            resetButton.clicked.connect(self._resetParams)
            layout.addWidget(resetButton)

        widget.setLayout(layout)
        return widget

    def _resetParams(self, checked=False):
        self._model.resetParams()


# PlainView + View to be refactored


class PlainView(QWidget):
    """Widget-based view for param.Model"""

    # Emited for every index when model data change
    dataChanged = Signal(object)

    def __init__(self, model=None, parent=None):
        super().__init__(parent)
        self._model = None
        self._rootIndex = QModelIndex()
        self._widgets = dict()
        self._customParamClass = dict()
        layout = QVBoxLayout()
        layout.setSpacing(8)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        if model is not None:
            self.setModel(model)

    def clearLayout(self):
        while self.layout().count():
            item = self.layout().takeAt(0)
            item.widget().deleteLater()

    def rootIndex(self):
        return self._rootIndex

    def setRootIndex(self, index):
        self._rootIndex = index
        self.draw()

    def model(self):
        return self._model

    def setModel(self, model):
        self._widgets = dict()
        self._model = model
        model.dataChanged.connect(self.onModelDataChange)
        self.setRootIndex(QModelIndex())

    def draw(self):
        self.clearLayout()
        self._populate(self._rootIndex, 0)

    def onInvalidValue(self, error):
        """Called when a field value was invalid"""
        QMessageBox.warning(self, "Warning", error)

    def addCustomParamWidget(self, param, _class):
        """The view for `param` will be an instance of `_class`"""
        self._customParamClass[id(param)] = _class
        self.draw()

    def onModelDataChange(self, topLeft, bottomRight, roles):
        """Send signal updates for all indices changed"""
        if bottomRight.parent() != topLeft.parent():
            raise RuntimeError("View: Model dataChanged items have different parents")
        parent = topLeft.parent()
        top = topLeft.row()
        bottom = bottomRight.row()
        for row in range(top, bottom + 1):
            index = self._model.index(row, 0, parent)
            self.dataChanged.emit(index)

    def onWidgetDataChange(self, index, data):
        """Update model"""
        if not self._model.setData(index, data):
            self.onInvalidValue(self._model.setDataError)
            self.sender().refreshData()
            self.sender().setFocus()
            # QTimer.singleShot(0, self.sender().setFocus)

    def _fieldWidget(self, index, field, depth=0):
        """Generate and return an appropriate widget for the field"""
        type_to_widget = {
            bool: BoolWidget,
            float: EntryWidget,
            str: EntryWidget,
            int: EntryWidget,
        }
        if id(field) in self._customParamClass:
            _class = self._customParamClass[id(field)]
            widget = _class(index, field, self)
        elif field.list:
            widget = ListWidget(index, field, self)
        elif field.type in type_to_widget.keys():
            widget = type_to_widget[field.type](index, field, self)
        else:
            widget = EntryWidget(index, field, self)
        widget.setProperty("depth", depth)
        return widget

    def _populate(self, index, depth=0):
        """Recursively populate and return a widget with params"""
        data = self._model.data(index, self._model.DataRole)

        if isinstance(data, Field):
            return self._fieldWidget(index, data, depth)

        if depth != 0:
            return None

        rows = self._model.rowCount(index)
        for row in range(0, rows):
            childIndex = self._model.index(row, 0, index)
            child = self._populate(childIndex, depth + 1)
            if child is not None:
                self.layout().addWidget(child)



# -----------------------------------------------------------------------------
# Commons - Utility classes for iTaxoTools modules
# Copyright (C) 2021  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""Graphical widgets and GUI elements"""


from PySide6 import QtCore, QtGui, QtSvg, QtWidgets

import re


class VectorPixmap(QtGui.QPixmap):
    """A colored vector pixmap"""

    def __init__(self, fileName, size=None, colormap=None):
        """
        Load an SVG resource file and replace colors according to
        provided dictionary `colormap`. Only fill and stroke is replaced.
        Also scales the pixmap if a QSize is provided.
        """
        data = self.loadAndMap(fileName, colormap)

        renderer = QtSvg.QSvgRenderer(data)
        size = renderer.defaultSize() if size is None else size
        super().__init__(size)
        self.fill(QtCore.Qt.transparent)
        painter = QtGui.QPainter(self)
        renderer.render(painter)
        painter.end()

    @staticmethod
    def loadAndMap(fileName, colormap):
        file = QtCore.QFile(fileName)
        if not file.open(QtCore.QIODevice.ReadOnly):
            raise FileNotFoundError("Vector resource not found: " + fileName)
        text = file.readAll().data().decode()
        file.close()

        if colormap is not None:
            regex = "(?P<color>"
            regex += "|".join(map(re.escape, colormap.keys())) + ")"
            text = re.sub(regex, lambda mo: colormap[mo.group("color")], text)

        return QtCore.QByteArray(text.encode())


class VectorIcon(QtGui.QIcon):
    """A colored vector icon"""

    def __init__(self, fileName, colormap_modes, size=None):
        """Create pixmaps with colormaps matching the dictionary modes"""
        super().__init__()
        for mode in colormap_modes.keys():
            pixmap = VectorPixmap(fileName, size=size, colormap=colormap_modes[mode])
            self.addPixmap(pixmap, mode)


class ScalingImage(QtWidgets.QLabel):
    """Keep aspect ratio, width adjusts with height"""

    def __init__(self, pixmap=None):
        """Remember given pixmap and ratio"""
        super().__init__()
        self.setScaledContents(False)
        self._polished = False
        self._logo = None
        self._ratio = 0
        if pixmap is not None:
            self.logo = pixmap

    @property
    def logo(self):
        return self._logo

    @logo.setter
    def logo(self, logo):
        self._logo = logo
        self._ratio = logo.width() / logo.height()
        self._scale()

    def _scale(self):
        """Create new pixmap to match new sizes"""
        if self._logo is None:
            return
        h = self.height()
        w = h * self._ratio
        self.setPixmap(
            self._logo.scaled(
                w, h, QtCore.Qt.KeepAspectRatio, QtCore.Qt.SmoothTransformation
            )
        )

    def minimumSizeHint(self):
        return QtCore.QSize(1, 1)

    def sizeHint(self):
        if self._polished is True and self._ratio != 0:
            h = self.height()
            return QtCore.QSize(h * self._ratio, h)
        else:
            return QtCore.QSize(1, 1)

    def resizeEvent(self, event):
        self._scale()
        super().resizeEvent(event)

    def event(self, ev):
        """Let sizeHint know that sizes are now real"""
        if ev.type() == QtCore.QEvent.PolishRequest:
            self._polished = True
            self.updateGeometry()
        return super().event(ev)



from itaxotools.common.bindings import Binder, Property, PropertyObject


class DummyObject(PropertyObject):
    dummy = Property(int, 0)


def test_simple_bind(qapp):
    a = DummyObject()
    b = DummyObject()

    binder = Binder()
    binder.bind(a.properties.dummy, b.properties.dummy)

    assert b.dummy == 0
    a.dummy = 42
    assert b.dummy == 42


def test_proxy_bind(qapp):
    a = DummyObject()
    b = DummyObject()

    binder = Binder()
    binder.bind(a.properties.dummy, b.properties.dummy, proxy=lambda x: x + 1)

    assert b.dummy == 1
    a.dummy = 42
    assert b.dummy == 43


def test_conditional_bind(qapp):
    a = DummyObject()
    b = DummyObject()

    binder = Binder()
    binder.bind(a.properties.dummy, b.properties.dummy, condition=lambda x: x > 10)

    assert b.dummy == 0

    a.dummy = 9
    assert b.dummy == 0

    a.dummy = 11
    assert b.dummy == 11


def test_binder_update(qapp):
    a = DummyObject()
    b = DummyObject()

    binder = Binder()
    binder.bind(a.properties.dummy, b.properties.dummy)

    a.dummy = 1
    assert b.dummy == 1

    b.dummy = 2
    assert b.dummy == 2

    binder.update()
    assert b.dummy == 1


def test_property_tags(qapp):
    class DummyTaggedObject(PropertyObject):
        x = Property(int, 0, tag="x")
        y = Property(int, 0, tag="y")
        z = Property(int, 0)

    a = DummyTaggedObject()

    assert a.properties.x.tag == "x"
    assert a.properties.y.tag == "y"
    assert a.properties.z.tag is None


def test_property_iteration(qapp):
    class DummyTaggedObject(PropertyObject):
        x = Property(int, 1)
        y = Property(int, 2)
        z = Property(int, 3)

    a = DummyTaggedObject()

    assert [x.key for x in a.properties] == ["x", "y", "z"]
    assert [x.value for x in a.properties] == [1, 2, 3]



# -----------------------------------------------------------------------------
# Commons - Utility classes for iTaxoTools modules
# Copyright (C) 2021-2023  Patmanidis Stefanos
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""Utility classes and functions"""


from typing import Callable, Generic, TypeVar

KeyType = TypeVar("KeyType")
DecoratedType = TypeVar("DecoratedType")


class DecoratorDict(Generic[KeyType, DecoratedType]):
    """
    Instances can be used as decorators that accept a single key as argument.
    Decorated functions can then be accessed by key.
    """

    def __init__(self):
        self.items = dict()

    def __call__(self, key: KeyType) -> Callable[[DecoratedType], DecoratedType]:
        def decorator(func: DecoratedType) -> DecoratedType:
            self.items[key] = func
            return func

        return decorator

    def __getitem__(self, key: KeyType) -> DecoratedType:
        if key not in self.items:
            raise Exception("")
        return self.items[key]

    def __contains__(self, key) -> bool:
        return key in self.items

    def __iter__(self) -> iter:
        return iter(self.items)

    def __len__(self) -> int:
        return len(self.items)


class AttrDict(dict):
    """Members can also be accessed as attributes"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    def __iter__(self):
        return iter(self.values())


class Guard:
    """Should probably use threading.Lock"""

    def __init__(self):
        self.locked = False

    def __enter__(self):
        self.locked = True

    def __exit__(self, type, value, trace):
        self.locked = False

    def __bool__(self):
        return self.locked


def override(f: DecoratedType) -> DecoratedType:
    return f


def type_convert(value, type, default):
    try:
        return type(value)
    except ValueError:
        return default


def human_readable_size(size):
    for unit in ["B", "KB", "MB", "GB"]:
        if size < 1000.0 or unit == "GB":
            break
        size /= 1000.0
    return f"{size:.2f} {unit}"


def human_readable_seconds(seconds):
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)

    segments = [
        f'{int(h)} hour{"s" if h >= 2 else ""}' if h else None,
        f'{int(m)} minute{"s" if m >= 2 else ""}' if m else None,
        f"{s:.2f} seconds" if s else None,
    ]
    segments = (x for x in segments if x)
    return ", ".join(segments)
